from _typeshed import Incomplete
from typing import Generic, TypeVar

from . import elements, roles
from .base import Options
from .operators import ColumnOperators

_T = TypeVar("_T")

class LambdaOptions(Options):
    enable_tracking: bool
    track_closure_variables: bool
    track_on: Incomplete
    global_track_bound_values: bool
    track_bound_values: bool
    lambda_cache: Incomplete

def lambda_stmt(
    lmb,
    enable_tracking: bool = ...,
    track_closure_variables: bool = ...,
    track_on: Incomplete | None = ...,
    global_track_bound_values: bool = ...,
    track_bound_values: bool = ...,
    lambda_cache: Incomplete | None = ...,
): ...

class LambdaElement(elements.ClauseElement):
    __visit_name__: str
    parent_lambda: Incomplete
    fn: Incomplete
    role: Incomplete
    tracker_key: Incomplete
    opts: Incomplete
    def __init__(self, fn, role, opts=..., apply_propagate_attrs: Incomplete | None = ...) -> None: ...
    def __getattr__(self, key: str): ...

class DeferredLambdaElement(LambdaElement):
    lambda_args: Incomplete
    def __init__(self, fn, role, opts=..., lambda_args=...) -> None: ...

class StatementLambdaElement(roles.AllowsLambdaRole, LambdaElement):
    def __add__(self, other): ...
    def add_criteria(
        self,
        other,
        enable_tracking: bool = ...,
        track_on: Incomplete | None = ...,
        track_closure_variables: bool = ...,
        track_bound_values: bool = ...,
    ): ...
    def spoil(self): ...

class NullLambdaStatement(roles.AllowsLambdaRole, elements.ClauseElement):
    __visit_name__: str
    def __init__(self, statement) -> None: ...
    def __getattr__(self, key: str): ...
    def __add__(self, other): ...
    def add_criteria(self, other, **kw): ...

class LinkedLambdaElement(StatementLambdaElement):
    role: Incomplete
    opts: Incomplete
    fn: Incomplete
    parent_lambda: Incomplete
    tracker_key: Incomplete
    def __init__(self, fn, parent_lambda, opts) -> None: ...

class AnalyzedCode:
    @classmethod
    def get(cls, fn, lambda_element, lambda_kw, **kw): ...
    track_bound_values: Incomplete
    track_closure_variables: Incomplete
    bindparam_trackers: Incomplete
    closure_trackers: Incomplete
    build_py_wrappers: Incomplete
    def __init__(self, fn, lambda_element, opts) -> None: ...

class NonAnalyzedFunction:
    closure_bindparams: Incomplete
    bindparam_trackers: Incomplete
    expr: Incomplete
    def __init__(self, expr) -> None: ...
    @property
    def expected_expr(self): ...

class AnalyzedFunction:
    analyzed_code: Incomplete
    fn: Incomplete
    closure_pywrappers: Incomplete
    tracker_instrumented_fn: Incomplete
    expr: Incomplete
    bindparam_trackers: Incomplete
    expected_expr: Incomplete
    is_sequence: Incomplete
    propagate_attrs: Incomplete
    closure_bindparams: Incomplete
    def __init__(self, analyzed_code, lambda_element, apply_propagate_attrs, fn) -> None: ...

class PyWrapper(ColumnOperators[_T], Generic[_T]):
    fn: Incomplete
    track_bound_values: Incomplete
    def __init__(
        self,
        fn,
        name,
        to_evaluate,
        closure_index: Incomplete | None = ...,
        getter: Incomplete | None = ...,
        track_bound_values: bool = ...,
    ) -> None: ...
    def __call__(self, *arg, **kw): ...
    def operate(self, op, *other, **kwargs): ...
    def reverse_operate(self, op, other, **kwargs): ...
    def __clause_element__(self): ...  # Field not always present.
    def __bool__(self) -> bool: ...
    def __nonzero__(self) -> bool: ...
    def __getattribute__(self, key: str): ...
    def __iter__(self): ...
    def __getitem__(self, key) -> ColumnOperators[_T]: ...

def insp(lmb): ...



from _typeshed import Incomplete

from .base import PARSE_AUTOCOMMIT as PARSE_AUTOCOMMIT, ColumnCollection as ColumnCollection, Executable as Executable
from .dml import Delete as Delete, Insert as Insert, Update as Update, UpdateBase as UpdateBase, ValuesBase as ValuesBase
from .elements import (
    BinaryExpression as BinaryExpression,
    BindParameter as BindParameter,
    BooleanClauseList as BooleanClauseList,
    Case as Case,
    Cast as Cast,
    ClauseElement as ClauseElement,
    ClauseList as ClauseList,
    CollectionAggregate as CollectionAggregate,
    ColumnClause as ColumnClause,
    ColumnElement as ColumnElement,
    Extract as Extract,
    False_ as False_,
    FunctionFilter as FunctionFilter,
    Grouping as Grouping,
    Label as Label,
    Null as Null,
    Over as Over,
    ReleaseSavepointClause as ReleaseSavepointClause,
    RollbackToSavepointClause as RollbackToSavepointClause,
    SavepointClause as SavepointClause,
    TextClause as TextClause,
    True_ as True_,
    Tuple as Tuple,
    TypeClause as TypeClause,
    TypeCoerce as TypeCoerce,
    UnaryExpression as UnaryExpression,
    WithinGroup as WithinGroup,
    _truncated_label as _truncated_label,
    between as between,
    collate as collate,
    literal as literal,
    literal_column as literal_column,
    not_ as not_,
    outparam as outparam,
    quoted_name as quoted_name,
)
from .functions import Function as Function, FunctionElement as FunctionElement, func as func, modifier as modifier
from .lambdas import LambdaElement as LambdaElement, StatementLambdaElement as StatementLambdaElement, lambda_stmt as lambda_stmt
from .operators import ColumnOperators as ColumnOperators, Operators as Operators, custom_op as custom_op
from .selectable import (
    CTE as CTE,
    LABEL_STYLE_DEFAULT as LABEL_STYLE_DEFAULT,
    LABEL_STYLE_DISAMBIGUATE_ONLY as LABEL_STYLE_DISAMBIGUATE_ONLY,
    LABEL_STYLE_NONE as LABEL_STYLE_NONE,
    LABEL_STYLE_TABLENAME_PLUS_COL as LABEL_STYLE_TABLENAME_PLUS_COL,
    Alias as Alias,
    AliasedReturnsRows as AliasedReturnsRows,
    CompoundSelect as CompoundSelect,
    Exists as Exists,
    FromClause as FromClause,
    FromGrouping as FromGrouping,
    GenerativeSelect as GenerativeSelect,
    HasCTE as HasCTE,
    HasPrefixes as HasPrefixes,
    HasSuffixes as HasSuffixes,
    Join as Join,
    Lateral as Lateral,
    ReturnsRows as ReturnsRows,
    ScalarSelect as ScalarSelect,
    Select as Select,
    Selectable as Selectable,
    SelectBase as SelectBase,
    Subquery as Subquery,
    TableClause as TableClause,
    TableSample as TableSample,
    TableValuedAlias as TableValuedAlias,
    TextAsFrom as TextAsFrom,
    TextualSelect as TextualSelect,
    Values as Values,
    subquery as subquery,
)
from .traversals import CacheKey as CacheKey
from .visitors import Visitable as Visitable

__all__ = [
    "Alias",
    "AliasedReturnsRows",
    "any_",
    "all_",
    "CacheKey",
    "ClauseElement",
    "ColumnCollection",
    "ColumnElement",
    "CompoundSelect",
    "Delete",
    "FromClause",
    "Insert",
    "Join",
    "Lateral",
    "LambdaElement",
    "StatementLambdaElement",
    "Select",
    "Selectable",
    "TableClause",
    "TableValuedAlias",
    "Update",
    "Values",
    "alias",
    "and_",
    "asc",
    "between",
    "bindparam",
    "case",
    "cast",
    "column",
    "custom_op",
    "cte",
    "delete",
    "desc",
    "distinct",
    "except_",
    "except_all",
    "exists",
    "extract",
    "func",
    "modifier",
    "collate",
    "insert",
    "intersect",
    "intersect_all",
    "join",
    "label",
    "lateral",
    "lambda_stmt",
    "literal",
    "literal_column",
    "not_",
    "null",
    "nulls_first",
    "nulls_last",
    "or_",
    "outparam",
    "outerjoin",
    "over",
    "select",
    "table",
    "text",
    "tuple_",
    "type_coerce",
    "quoted_name",
    "union",
    "union_all",
    "update",
    "quoted_name",
    "within_group",
    "Subquery",
    "TableSample",
    "tablesample",
    "values",
]

all_: Incomplete
any_: Incomplete
and_: Incomplete
alias: Incomplete
tablesample: Incomplete
lateral: Incomplete
or_: Incomplete
bindparam: Incomplete
select: Incomplete
text: Incomplete
table: Incomplete
column: Incomplete
over: Incomplete
within_group: Incomplete
label: Incomplete
case: Incomplete
cast: Incomplete
cte: Incomplete
values: Incomplete
extract: Incomplete
tuple_: Incomplete
except_: Incomplete
except_all: Incomplete
intersect: Incomplete
intersect_all: Incomplete
union: Incomplete
union_all: Incomplete
exists: Incomplete
nulls_first: Incomplete
nullsfirst: Incomplete
nulls_last: Incomplete
nullslast: Incomplete
asc: Incomplete
desc: Incomplete
distinct: Incomplete
type_coerce: Incomplete
true: Incomplete
false: Incomplete
null: Incomplete
join: Incomplete
outerjoin: Incomplete
insert: Incomplete
update: Incomplete
delete: Incomplete
funcfilter: Incomplete



from _typeshed import Incomplete
from typing import Any

from ..util import HasMemoized
from .base import Executable, Generative
from .elements import BinaryExpression, ColumnElement, NamedColumn
from .selectable import FromClause, TableValuedAlias
from .visitors import TraversibleType

def register_function(identifier, fn, package: str = ...) -> None: ...

class FunctionElement(Executable, ColumnElement[Any], FromClause, Generative):  # type: ignore[misc]
    packagenames: Incomplete
    clause_expr: Incomplete
    def __init__(self, *clauses, **kwargs) -> None: ...
    def scalar_table_valued(self, name, type_: Incomplete | None = ...): ...
    def table_valued(self, *expr, **kw): ...
    def column_valued(self, name: str | None = ..., joins_implicitly: bool = ...): ...
    @property
    def columns(self): ...
    @property
    def exported_columns(self): ...
    @HasMemoized.memoized_attribute
    def clauses(self): ...
    def over(
        self,
        partition_by: Incomplete | None = ...,
        order_by: Incomplete | None = ...,
        rows: Incomplete | None = ...,
        range_: Incomplete | None = ...,
    ): ...
    def within_group(self, *order_by): ...
    def filter(self, *criterion): ...
    def as_comparison(self, left_index, right_index): ...
    def within_group_type(self, within_group) -> None: ...
    def alias(self, name: str | None = ..., joins_implicitly: bool = ...) -> TableValuedAlias: ...  # type: ignore[override]
    def select(self): ...
    def scalar(self): ...
    def execute(self): ...
    def self_group(self, against: Incomplete | None = ...): ...
    @property
    def entity_namespace(self): ...

class FunctionAsBinary(BinaryExpression):
    sql_function: Incomplete
    left_index: Incomplete
    right_index: Incomplete
    operator: Incomplete
    type: Incomplete
    negate: Incomplete
    modifiers: Incomplete
    def __init__(self, fn, left_index, right_index) -> None: ...
    @property
    def left(self): ...
    @left.setter
    def left(self, value) -> None: ...
    @property
    def right(self): ...
    @right.setter
    def right(self, value) -> None: ...

class ScalarFunctionColumn(NamedColumn):
    __visit_name__: str
    is_literal: bool
    table: Incomplete
    fn: Incomplete
    name: Incomplete
    type: Incomplete
    def __init__(self, fn, name, type_: Incomplete | None = ...) -> None: ...

class _FunctionGenerator:
    opts: Incomplete
    def __init__(self, **opts) -> None: ...
    def __getattr__(self, name: str): ...
    def __call__(self, *c, **kwargs): ...

func: Incomplete
modifier: Incomplete

class Function(FunctionElement):
    __visit_name__: str
    type: Incomplete
    packagenames: Incomplete
    name: Incomplete
    def __init__(self, name, *clauses, **kw) -> None: ...

class _GenericMeta(TraversibleType):
    def __init__(cls, clsname, bases, clsdict) -> None: ...

class GenericFunction:
    name: Incomplete
    identifier: Incomplete
    coerce_arguments: bool
    inherit_cache: bool
    packagenames: Incomplete
    clause_expr: Incomplete
    type: Incomplete
    def __init__(self, *args, **kwargs) -> None: ...

class next_value(GenericFunction):
    type: Incomplete
    name: str
    sequence: Incomplete
    def __init__(self, seq, **kw) -> None: ...
    def compare(self, other, **kw): ...

class AnsiFunction(GenericFunction):
    inherit_cache: bool
    def __init__(self, *args, **kwargs) -> None: ...

class ReturnTypeFromArgs(GenericFunction):
    inherit_cache: bool
    def __init__(self, *args, **kwargs) -> None: ...

class coalesce(ReturnTypeFromArgs):
    inherit_cache: bool

class max(ReturnTypeFromArgs):
    inherit_cache: bool

class min(ReturnTypeFromArgs):
    inherit_cache: bool

class sum(ReturnTypeFromArgs):
    inherit_cache: bool

class now(GenericFunction):
    type: Incomplete
    inherit_cache: bool

class concat(GenericFunction):
    type: Incomplete
    inherit_cache: bool

class char_length(GenericFunction):
    type: Incomplete
    inherit_cache: bool
    def __init__(self, arg, **kwargs) -> None: ...

class random(GenericFunction):
    inherit_cache: bool

class count(GenericFunction):
    type: Incomplete
    inherit_cache: bool
    def __init__(self, expression: Incomplete | None = ..., **kwargs) -> None: ...

class current_date(AnsiFunction):
    type: Incomplete
    inherit_cache: bool

class current_time(AnsiFunction):
    type: Incomplete
    inherit_cache: bool

class current_timestamp(AnsiFunction):
    type: Incomplete
    inherit_cache: bool

class current_user(AnsiFunction):
    type: Incomplete
    inherit_cache: bool

class localtime(AnsiFunction):
    type: Incomplete
    inherit_cache: bool

class localtimestamp(AnsiFunction):
    type: Incomplete
    inherit_cache: bool

class session_user(AnsiFunction):
    type: Incomplete
    inherit_cache: bool

class sysdate(AnsiFunction):
    type: Incomplete
    inherit_cache: bool

class user(AnsiFunction):
    type: Incomplete
    inherit_cache: bool

class array_agg(GenericFunction):
    type: Incomplete
    inherit_cache: bool
    def __init__(self, *args, **kwargs) -> None: ...

class OrderedSetAgg(GenericFunction):
    array_for_multi_clause: bool
    inherit_cache: bool
    def within_group_type(self, within_group): ...

class mode(OrderedSetAgg):
    inherit_cache: bool

class percentile_cont(OrderedSetAgg):
    array_for_multi_clause: bool
    inherit_cache: bool

class percentile_disc(OrderedSetAgg):
    array_for_multi_clause: bool
    inherit_cache: bool

class rank(GenericFunction):
    type: Incomplete
    inherit_cache: bool

class dense_rank(GenericFunction):
    type: Incomplete
    inherit_cache: bool

class percent_rank(GenericFunction):
    type: Incomplete
    inherit_cache: bool

class cume_dist(GenericFunction):
    type: Incomplete
    inherit_cache: bool

class cube(GenericFunction):
    inherit_cache: bool

class rollup(GenericFunction):
    inherit_cache: bool

class grouping_sets(GenericFunction):
    inherit_cache: bool
